@@ -1,17 +1,21 @@
 package HTML::TreeBuilder::XPath;
 
+use List::Util qw( first);
+
 use strict;
 use warnings;
 
 use vars qw($VERSION);
 
-$VERSION = '0.08';
+$VERSION = '0.10';
+
+my %CHAR2DEFAULT_ENT= ( '&' => '&amp;', '<' => '&lt;', '>' => '&gt;', '"' => '&quot;');
+my %NUM2DEFAULT_ENT= ( '38' => 'amp', '60' => 'lt', '62' => 'gt', '"' => '&quot;');
 
 package HTML::TreeBuilder::XPath;
 
 use base( 'HTML::TreeBuilder');
 
-
 package HTML::TreeBuilder::XPath::Node;
 
 sub isElementNode   { 0 }
@@ -26,6 +30,13 @@ sub getChildNodes { return wantarray ? () : []; }
 sub getFirstChild { return undef; }
 sub getLastChild { return undef; }
 
+# need to do a complete look_down each time, as the id could have been changed 
+# without any object being involved, hence without a potential cache being
+# up to date
+sub getElementById 
+  { my ($self, $id) = @_;
+    return scalar $self->look_down( id => $id);
+  }
 
 sub to_number { return XML::XPathEngine::Number->new( shift->getValue); }
 
@@ -37,7 +48,7 @@ sub cmp
 
     # easy cases
     return  0 if( $a == $b);    
-    return 1 if( $a->is_inside($b)); # a starts after b 
+    return  1 if( $a->is_inside($b)); # a starts after b 
     return -1 if( $b->is_inside($a)); # a starts before b
 
     # lineage does not include the element itself
@@ -48,8 +59,8 @@ sub cmp
     unless( $a_pile[-1] == $b_pile[-1]) 
       { warn "2 nodes not in the same pile: ", ref( $a), " - ", ref( $b), "\n"; 
         print "a: ", $a->string_value, "\nb: ", $b->string_value, "\n";
+        return undef;
       }
-    return undef unless( $a_pile[-1] == $b_pile[-1]);
 
     # find the first non common ancestors (they are siblings)
     my $a_anc= pop @a_pile;
@@ -63,7 +74,7 @@ sub cmp
     if( defined( $a_anc->{_rank}) && defined( $b_anc->{_rank}))
       { return $a_anc->{_rank} <=> $b_anc->{_rank}; }
     else
-      { warn "no rank found";
+      {
         # from there move left and right and figure out the order
         my( $a_prev, $a_next, $b_prev, $b_next)= ($a_anc, $a_anc, $b_anc, $b_anc);
         while()
@@ -97,13 +108,16 @@ use XML::XPathEngine;
     }
 }
 
-sub findnodes           { my( $elt, $path)= @_; return xp->findnodes(           $path, $elt); }
-sub findnodes_as_string { my( $elt, $path)= @_; return xp->findnodes_as_string( $path, $elt); }
-sub findvalue           { my( $elt, $path)= @_; return xp->findvalue(           $path, $elt); }
-sub exists              { my( $elt, $path)= @_; return xp->exists(              $path, $elt); }
-sub find_xpath          { my( $elt, $path)= @_; return xp->find(                $path, $elt); }
-sub matches             { my( $elt, $path)= @_; return xp->matches( $elt, $path, $elt); }
-sub set_namespace       { my $elt= shift; xp->new->set_namespace( @_); }
+
+sub findnodes            { my( $elt, $path)= @_; return xp->findnodes(            $path, $elt); }
+sub findnodes_as_string  { my( $elt, $path)= @_; return xp->findnodes_as_string(  $path, $elt); }
+sub findnodes_as_strings { my( $elt, $path)= @_; return xp->findnodes_as_strings( $path, $elt); }
+sub findvalue            { my( $elt, $path)= @_; return xp->findvalue(            $path, $elt); }
+sub findvalues           { my( $elt, $path)= @_; return xp->findvalues(           $path, $elt); }
+sub exists               { my( $elt, $path)= @_; return xp->exists(               $path, $elt); }
+sub find_xpath           { my( $elt, $path)= @_; return xp->find(                 $path, $elt); }
+sub matches              { my( $elt, $path)= @_; return xp->matches( $elt, $path, $elt);        }
+sub set_namespace        { my $elt= shift; xp->new->set_namespace( @_); }
 
 sub getRootNode
   { my $elt= shift;
@@ -198,6 +212,187 @@ sub _child_as_object
     return $elt_or_text;
   }
 
+sub toString { return shift->as_XML( @_); }
+
+# produces better looking XML
+{
+  no warnings 'redefine'; 
+  sub as_XML_compact
+    { my( $node, $opt)= @_;
+      my $name = $node->{'_tag'};
+      if( $name eq '~literal')     { return _xml_escape_text( $node->{text});                     }
+
+      if( $name eq '~declaration') { return '<!'   . _xml_escape_text( $node->{text})    . ">";   }
+      if( $name eq '~pi')          { return '<?'   . _xml_escape_text( $node->{text})    . '?>';  }
+      if( $name eq '~comment')     { return '<!--' . _xml_escape_comment( $node->{text}) . '-->'; }
+
+      my $lc_name= lc $name;
+
+      my $xml= $node->_start_tag;
+
+      if( $HTML::Tagset::isCDATA_Parent{$lc_name})
+        { my $content= $node->{_content} || '';
+          if( ref $content eq 'ARRAY' || $content->isa( 'ARRAY'))
+            { $xml .= _xml_escape_cdata( join( '', @$content), $opt); }
+          else
+            { $xml .= $content; }
+        }
+      else
+        { # start tag
+          foreach my $child ($node->content_list) 
+            { if( ref $child) { $xml .= $child->as_XML_compact(); }
+              else            { $xml .=  _xml_escape_text( $child); }
+            }
+        }
+          $xml.= "</$name>" unless $HTML::Tagset::emptyElement{$lc_name};
+      return $xml;
+    }
+}
+
+    
+
+{ my %phrase_name;    # all phrase tags, + literals (those are not indented)
+  my %extra_newline;  # tags that get an extra newline before the end tag
+  my $default_indent; # 2 spaces, change with the 'indent' option
+  BEGIN 
+    { %phrase_name= %HTML::Tagset::isPhraseMarkup;
+      $phrase_name{'~literal'}= 1;
+      $default_indent= '  ';
+      %extra_newline= map { $_ => 1 } qw(html head body script div table tbody thead tfoot tr form dl ol ul);
+    }
+
+  sub as_XML_indented
+    { my( $node, $opt)= @_;
+
+
+      my $name = $node->{'_tag'};
+      my $lc_name= lc $name;
+
+      if( $name eq '~literal')     { return _xml_escape_text( $node->{text});                     }
+      if( $name eq '~declaration') { return '<!'   . _xml_escape_text( $node->{text})    . ">\n"; }
+
+
+      if( $name eq '~pi')          { return '<?'   . _xml_escape_text( $node->{text})    . '?>';  }
+      if( $name eq '~comment')     { return '<!--' . _xml_escape_comment( $node->{text}) . '-->'; }
+      
+      my $xml;
+      my $pre_tag_indent='';
+      if(!$phrase_name{$lc_name}) { $pre_tag_indent=  "\n" . ($opt->{indent} || $default_indent) x ($opt->{indent_level}||0); }
+      if( $opt->{indent_level}) { $xml .= $pre_tag_indent; }
+
+      $xml.= $node->_start_tag();
+
+      my $content='';
+
+      if( $HTML::Tagset::isCDATA_Parent{$lc_name})
+        { my $content= $node->{_content} || '';
+          if( ref $content && (ref $content eq 'ARRAY' || $content->isa( 'ARRAY') ))
+            { $content= _xml_escape_cdata( join( '', @$content), $opt); }
+        }
+      else
+        { 
+          my %child_opt= %$opt;
+          $child_opt{indent_level}++;
+          foreach my $child ($node->content_list) 
+            { if( ref $child) { $content .= $child->as_XML_indented( \%child_opt ); }
+              else            { $content .=  _xml_escape_text( $child);             }
+            }
+        }
+      $xml .= $content;
+
+      if( $extra_newline{$lc_name} && $content ne '' ) { $xml.= $pre_tag_indent; }
+      $xml.= "</$name>" unless $HTML::Tagset::emptyElement{$lc_name};
+      $xml .="\n" if( !$opt->{indent_level});
+     
+      return $xml;
+    }
+}
+
+sub _start_tag
+  { my( $node)= @_;
+    my $name = $node->{'_tag'};
+    my $start_tag.= "<$name";
+    foreach my $att_name (sort keys %$node) 
+      { next if( (!length $att_name) ||  ($att_name=~ m{^_}) || ($att_name eq '/') );
+        my $well_formed_att_name= well_formed_name( $att_name);
+        $start_tag .= qq{ $well_formed_att_name="} . _xml_escape_attribute_value( $node->{$att_name}) . qq{"};
+      }
+    $start_tag.= $HTML::Tagset::emptyElement{lc $name} ? " />" : ">";
+    return $start_tag;
+  }
+
+sub well_formed_name
+  { my( $name)= @_;
+    $name=~ s{[^\w:_-]+}{_}g;
+    if( $name=~ m{^\d}) { $name= "a$name"; }
+    return $name;
+  }
+
+sub _indent_level
+  { my( $node)= @_;
+    my $level= scalar grep { !$HTML::Tagset::isPhraseMarkup{lc $_->{_tag}} } $node->lineage;
+    return $level;
+  }
+
+{ my( $indent, %extra_newline, $nl);
+  BEGIN 
+    { $indent= '  '; 
+      $nl= "\n";
+      %extra_newline= map { $_ => 1 } qw(html head body script div table tr form ol ul);
+    }
+   
+  sub indents
+    { my( $opt, $name)= @_;
+      my $indents= { pre_start_tag => '', post_start_tag => '', pre_end_tag => '', post_end_tag => ''};
+      if( $opt->{indented})
+        { my $indent_level= $opt->{indent_level};
+          my $wrapping_nl= $nl; 
+          if( !defined( $indent_level)) { $indent_level = 0; $wrapping_nl= ''; }
+          if( $HTML::Tagset::isKnown{lc $name} && !$HTML::Tagset::isPhraseMarkup{lc $name} && $indent_level > 0) 
+            { $indents->{pre_start_tag}= $wrapping_nl . ($indent x $indent_level); }
+          if( $extra_newline{lc $name})
+            { $indents->{post_start_tag}= $nl; 
+              $indents->{pre_end_tag}= $nl . ($indent x $indent_level);
+            }
+          if( $indent_level == 0) 
+            { $indents->{post_end_tag} = $wrapping_nl; }
+        }
+      return $indents;
+    }
+}
+
+    
+sub _xml_escape_attribute_value
+  { my( $text)= @_;
+    $text=~ s{([&<>"])}{$CHAR2DEFAULT_ENT{$1}}g; # escape also quote, as it is the attribute separator
+    return $text;
+  }
+
+sub _xml_escape_text
+  { my( $text)= @_;
+    $text=~ s{([&<>])}{$CHAR2DEFAULT_ENT{$1}}g;
+    return $text;
+  }
+
+sub _xml_escape_comment
+  { my( $text)= @_;
+    $text=~ s{([&<>])}{$CHAR2DEFAULT_ENT{$1}}g;
+    $text=~ s{--}{-&#45;}g; # can't have double --'s in XML comments
+    return $text;
+  }
+
+sub _xml_escape_cdata
+  { my( $text, $opt)= @_;
+    if( $opt->{force_escape_cdata} || $text=~ m{[<&]})
+      { $text=~ s{^\s*\Q<![CDATA[}{}s;
+        $text=~ s{\Q]]>\E\s*$}{}s;
+        $text=~ s{]]>}{]]&#62;}g; # can't have]]> in CDATA
+        $text=  "<![CDATA[$text]]>";
+      }
+    return $text;
+  }
+
+
 package HTML::TreeBuilder::XPath::TextNode;
 
 use base 'HTML::TreeBuilder::XPath::Node';
@@ -206,7 +401,19 @@ sub getParentNode { return shift->{_parent};    }
 sub getValue      { return shift->{_content};   }
 sub isTextNode    { return 1;                   }
 sub getAttributes { return wantarray ? () : []; }
-sub as_XML        { return shift->getValue;     }
+
+# similar to HTML::Element as_XML
+sub as_XML
+  { my( $node, $entities)= @_;
+    my $content= $node->{_content};
+    if( $node->{_parent} && $node->{_parent}->{_tag} eq 'script')
+      { $content=~ s{(&\w+;)}{HTML::Entities::decode($1)}eg; }
+    else
+      { $content= HTML::Element::_xml_escape_text($content); }
+    return $content;
+  }
+*as_XML_compact  = *as_XML;
+*as_XML_indented = *as_XML;
 
 
 sub getPreviousSibling
@@ -258,7 +465,7 @@ sub getLocalName    { (my $name= $_[0]->{_name}) =~ s{^.*:}{}; $name; }
 sub string_value    { return $_[0]->{_value}; }
 sub to_number       { return XML::XPathEngine::Number->new( $_[0]->{_value}); }
 sub isAttributeNode { 1 }
-sub toString        { return qq{$_[0]->{_name}="$_[0]->{_value}"}; }
+sub toString        { return qq{ $_[0]->{_name}="$_[0]->{_value}"}; }
 
 # awfully inefficient, but hopefully this is called only for weird (read test-case) queries
 sub getPreviousSibling
@@ -311,6 +518,10 @@ sub getParentNode   { return (); }
 sub getChildNodes   { my @content= ( $_[0]->{_root}); return wantarray ? @content : \@content; }
 sub getAttributes   { return []        }
 sub isDocumentNode  { return 1         }
+sub getRootNode     { return $_[0]     }
+sub getName         { return           }
+sub getNextSibling  { return           }
+sub getPreviousSibling { return        }
 
 # added to provide element-like methods to root, for use by cmp
 sub lineage {  return ($_[0]); }
@@ -352,7 +563,11 @@ In scalar context returns an C<Tree::XPathEngine::NodeSet> object.
 
 =head2 findnodes_as_string ($path)
 
-Returns the text values of the nodes
+Returns the text values of the nodes, as one string.
+
+=head2 findnodes_as_strings ($path)
+
+Returns a list of the values of the result nodes. 
 
 =head2 findvalue ($path)
 
@@ -364,6 +579,12 @@ for each of the objects stringification is overloaded, so you can just
 print the value found, or manipulate it in the ways you would a normal
 perl value (e.g. using regular expressions).
 
+=head2 findvalues ($path)
+
+Returns the values of the matching nodes as a list. This is mostly the same
+as findnodes_as_strings, except that the elements of the list are objects
+(with overloaded stringification) instead of plain strings.
+
 =head2 exists ($path)
 
 Returns true if the given path exists.
@@ -382,7 +603,20 @@ return something - and you can use ->isa() to find out what it returned. If
 you need to check how many nodes it found you should check $nodeset->size.
 See L<XML::XPathEngine::NodeSet>.
 
+=head2 as_XML_compact
 
+HTML::TreeBuilder's C<as_XML> output is not really nice to look at, so
+I added a new method, that can be used as a simple replacement for it. 
+It escapes only the '<', '>' and '&' (plus '"' in attribute values), and
+wraps CDATA elements in CDATA sections.
+
+Note that the XML is actually not garanteed to be valid at this point. Nothing
+is done about the encoding of the string. Patches or just ideas of how it could
+work are welcome.
+
+=head2 as_XML_indented 
+
+Same as as_XML, except that the output is indented.
 
 =head1 SEE ALSO
 
@@ -392,7 +626,7 @@ L<XML::XPathEngine>
 
 =head1 AUTHOR
 
-Michel Rodriguez, E<lt>imirod@cpan.orgE<gt>
+Michel Rodriguez, E<lt>mirod@cpan.orgE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
